@@ -1,5 +1,5 @@
 #!/usr/bin/env python
-# Copyright (C) 2013-2018  Vincent Pelletier <plr.vincent@gmail.com>
+# Copyright (C) 2013-2020  Vincent Pelletier <plr.vincent@gmail.com>
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -39,8 +39,12 @@ Statistic profiling:
 >>>     # Code to profile
 >>> prof.print_stats()
 """
-
+from __future__ import print_function, division, absolute_import
 from collections import defaultdict, deque
+from email.mime.multipart import MIMEMultipart
+from email.mime.text import MIMEText
+from email.mime.application import MIMEApplication
+from email.encoders import encode_quopri
 from functools import partial, wraps
 # Note: use time, not clock.
 # Clock, at least on linux, ignores time not spent executing code
@@ -97,6 +101,8 @@ class BaseLineIterator(object):
         self._lineno += 1
         return lineno, self._getline(self._filename, lineno, self._global_dict)
 
+    next = __next__ # BBB 2.7
+
 if sys.version_info < (3, ):
     import codecs
     # Find coding specification (see PEP-0263)
@@ -128,13 +134,22 @@ if sys.version_info < (3, ):
             # else, first line is unicode.
 
         def __next__(self):
-            lineno, line = next(super(LineIterator, self))
+            lineno, line = super(LineIterator, self).__next__()
             if self._encoding:
                 line = line.decode(self._encoding, errors='replace')
             return lineno, line
+
+        next = __next__ # BBB
+
+    def iterframes(current_frames):
+        return current_frames().iteritems()
 else:
     # getline returns unicode objects, nothing to do
     LineIterator = BaseLineIterator
+    unicode = basestring = str
+
+    def iterframes(current_frames):
+        return current_frames().items()
 
 if platform.system() == 'Windows':
     quoteCommandline = windows_list2cmdline
@@ -326,22 +341,22 @@ class LocalDescriptor(threading.local):
             pass
 
 _ANNOTATE_HEADER = \
-    '%6s|%10s|' \
-    '%13s|%13s|%7s|' \
-    'Source code' % (
-        'Line #', 'Hits',
-        'Time', 'Time per hit', '%',
+    u'%6s|%10s|' \
+    u'%13s|%13s|%7s|' \
+    u'Source code' % (
+        u'Line #', u'Hits',
+        u'Time', u'Time per hit', u'%',
     )
-_ANNOTATE_HORIZONTAL_LINE = ''.join(x == '|' and '+' or '-'
+_ANNOTATE_HORIZONTAL_LINE = u''.join(x == u'|' and u'+' or u'-'
     for x in _ANNOTATE_HEADER)
 _ANNOTATE_FORMAT = \
-    '%(lineno)6i|%(hits)10i|' \
-    '%(time)13g|%(time_per_hit)13g|%(percent)6.2f%%|' \
-    '%(line)s'
+    u'%(lineno)6i|%(hits)10i|' \
+    u'%(time)13g|%(time_per_hit)13g|%(percent)6.2f%%|' \
+    u'%(line)s'
 _ANNOTATE_CALL_FORMAT = \
-    '(call)|%(hits)10i|' \
-    '%(time)13g|%(time_per_hit)13g|%(percent)6.2f%%|' \
-    '# %(callee_file)s:%(callee_line)s %(callee_name)s'
+    u'(call)|%(hits)10i|' \
+    u'%(time)13g|%(time_per_hit)13g|%(percent)6.2f%%|' \
+    u'# %(callee_file)s:%(callee_line)s %(callee_name)s'
 
 def _initStack():
     # frame_time: when current frame execution started/resumed last
@@ -494,7 +509,7 @@ class ProfileBase(object):
     def _getFileNameList(self, filename, may_sort=True):
         if filename is None:
             filename = self.getFilenameSet()
-        elif isinstance(filename, str):
+        elif isinstance(filename, basestring):
             return [filename]
         if may_sort:
             try:
@@ -508,6 +523,103 @@ class ProfileBase(object):
                 )
         return filename
 
+    def _iterOutFiles(self, filename=None, commandline=None):
+        """
+        Yields path, data, mimetype for each file involved on or produced by
+        profiling.
+        """
+        out = io.StringIO()
+        self.callgrind(
+            out,
+            filename=filename,
+            commandline=commandline,
+            relative_path=True,
+        )
+        yield (
+            'cachegrind.out.pprofile',
+            out.getvalue(),
+            'application/x-kcachegrind',
+        )
+        for name in self._getFileNameList(filename, may_sort=False):
+            lines = ''.join(self._iterRawFile(name))
+            if lines:
+                if isinstance(lines, unicode):
+                    lines = lines.encode('utf-8')
+                yield (
+                    _relpath(name),
+                    lines,
+                    'text/x-python',
+                )
+
+    def getCallgrindMIME(self, out, filename=None, commandline=None, relative_path=True):
+        """
+        Write to "out" a mime-multipart representation of:
+        - callgrind profiling statistics (cachegrind.out.pprofile)
+        - all involved python code, including Python Scripts without hierarchy
+          (the rest)
+        and return its mimetype.
+        To unpack resulting file, see "unpack a MIME message" in
+          http://docs.python.org/2/library/email-examples.html
+        Or get demultipart from
+          https://pypi.python.org/pypi/demultipart
+
+        relative_path:
+          Ignored.
+        """
+        _ = relative_path
+        result = MIMEMultipart()
+        base_type_dict = {
+            'application': MIMEApplication,
+            'text': MIMEText,
+        }
+        encoder_dict = {
+            'application/x-kcachegrind': encode_quopri,
+            'text/x-python': 'utf-8',
+            'text/plain': 'utf-8',
+        }
+        for path, data, mimetype in self._iterOutFiles(
+            filename=filename,
+            commandline=commandline,
+        ):
+            base_type, sub_type = mimetype.split('/')
+            chunk = base_type_dict[base_type](
+                data,
+                sub_type,
+                encoder_dict.get(mimetype),
+            )
+            chunk.add_header(
+                'Content-Disposition',
+                'attachment',
+                filename=path,
+            )
+            result.attach(chunk)
+        out.write(result.as_string())
+        return result['content-type']
+
+    def getCallgrindZip(self, out=None, filename=None, commandline=None, relative_path=True):
+        """
+        Write to "out" a serialised zip archive containing:
+        - callgrind profiling statistics (cachegrind.out.pprofile)
+        - all involved python code, including Python Scripts without hierarchy
+          (the rest)
+        as a byte string and the "application/zip" mimetype.
+
+        relative_path:
+          Ignored.
+        """
+        _ = relative_path
+        with zipfile.ZipFile(
+            out,
+            mode='w',
+            compression=zipfile.ZIP_DEFLATED,
+        ) as outfile:
+            for path, data, _ in self._iterOutFiles(
+                filename=filename,
+                commandline=commandline,
+            ):
+                outfile.writestr(path, data)
+        return 'application/zip'
+
     def callgrind(self, out, filename=None, commandline=None, relative_path=False):
         """
         Dump statistics in callgrind format.
@@ -519,6 +631,11 @@ class ProfileBase(object):
         Time unit: microsecond (1e-6 second).
         out (file-ish opened for writing)
             Destination of callgrind profiling data.
+            Encoding should be chosen to be able to represent characters in
+            filesystem path and (when provided) commandline.
+            Setting an encoding error handler other than "raise" (default)
+            will likely result in profiling results being unable to locate
+            code for annotation.
         filename (str, collection of str)
             If provided, dump stats for given source file(s) only.
             By default, list for all known files.
@@ -531,13 +648,13 @@ class ProfileBase(object):
             profiling result, so kcachegrind does not look in system-wide
             files which may not match with profiled code.
         """
-        print('# callgrind format', file=out)
-        print('version: 1', file=out)
-        print('creator: pprofile', file=out)
-        print('event: usphit :microseconds/hit', file=out)
-        print('events: hits microseconds usphit', file=out)
+        print(u'# callgrind format', file=out)
+        print(u'version: 1', file=out)
+        print(u'creator: pprofile', file=out)
+        print(u'event: usphit :microseconds/hit', file=out)
+        print(u'events: hits microseconds usphit', file=out)
         if commandline is not None:
-            print('cmd:', commandline, file=out)
+            print(u'cmd:', commandline, file=out)
         file_dict = self._mergeFileTiming()
         if relative_path:
             convertPath = _relpath
@@ -569,7 +686,7 @@ class ProfileBase(object):
                 return name
         for current_file in self._getFileNameList(filename, may_sort=False):
             file_timing = file_dict[current_file]
-            print('fl=%s' % convertPath(current_file), file=out)
+            print(u'fl=%s' % convertPath(current_file), file=out)
             # When a local callable is created an immediately executed, this
             # loop would start a new "fn=" section but would not end it before
             # emitting "cfn=" lines, making the callee appear as not being
@@ -594,12 +711,12 @@ class ProfileBase(object):
                     getCodeName(current_file, caller)
                 ][lineno][1]
                 append = func_call_list.append
-                append('cfl=' + convertPath(callee_file))
-                append('cfn=' + getCodeName(callee_file, callee))
-                append('calls=%i %i' % (call_hits, callee.co_firstlineno))
-                append('%i %i %i %i' % (lineno, call_hits, call_ticks, call_ticks // call_hits))
+                append(u'cfl=' + convertPath(callee_file))
+                append(u'cfn=' + getCodeName(callee_file, callee))
+                append(u'calls=%i %i' % (call_hits, callee.co_firstlineno))
+                append(u'%i %i %i %i' % (lineno, call_hits, call_ticks, call_ticks // call_hits))
             for func_name, line_dict in func_dict.items():
-                print('fn=%s' % func_name, file=out)
+                print(u'fn=%s' % func_name, file=out)
                 for lineno, (func_hit_list, func_call_list) in sorted(line_dict.items()):
                     if func_hit_list:
                         # Multiple function objects may "reside" on the same
@@ -608,7 +725,7 @@ class ProfileBase(object):
                         hits = sum(x for x, _ in func_hit_list)
                         ticks = sum(x for _, x in func_hit_list)
                         print(
-                            '%i %i %i %i' % (
+                            u'%i %i %i %i' % (
                                 lineno,
                                 hits,
                                 ticks,
@@ -626,6 +743,9 @@ class ProfileBase(object):
         Time unit: second.
         out (file-ish opened for writing)
             Destination of annotated sources.
+            Encoding and encoding error handling should be chosen so as to be
+            able to represent all characters present in source code (ex: utf-8,
+            or ascii with "replace" error handler).
         filename (str, collection of str)
             If provided, dump stats for given source file(s) only.
             If unordered collection, it will get sorted by decreasing total
@@ -640,8 +760,8 @@ class ProfileBase(object):
         file_dict = self._mergeFileTiming()
         total_time = self.total_time
         if commandline is not None:
-            print('Command line:', commandline, file=out)
-        print('Total duration: %gs' % total_time, file=out)
+            print(u'Command line:', commandline, file=out)
+        print(u'Total duration: %gs' % total_time, file=out)
         if not total_time:
             return
         def percent(value, scale):
@@ -652,8 +772,8 @@ class ProfileBase(object):
             file_timing = file_dict[name]
             file_total_time = file_timing.getTotalTime()
             call_list_by_line = file_timing.getCallListByLine()
-            print('File: %s' % name, file=out)
-            print('File duration: %gs (%.2f%%)' % (file_total_time,
+            print(u'File: %s' % name, file=out)
+            print(u'File duration: %gs (%.2f%%)' % (file_total_time,
                 percent(file_total_time, total_time)), file=out)
             print(_ANNOTATE_HEADER, file=out)
             print(_ANNOTATE_HORIZONTAL_LINE, file=out)
@@ -667,12 +787,12 @@ class ProfileBase(object):
                     break
                 hits, duration = file_timing.getHitStatsFor(lineno)
                 print(_ANNOTATE_FORMAT % {
-                    'lineno': lineno,
-                    'hits': hits,
-                    'time': duration,
-                    'time_per_hit': duration / hits if hits else 0,
-                    'percent': percent(duration, total_time),
-                    'line': (line or '').rstrip(),
+                    u'lineno': lineno,
+                    u'hits': hits,
+                    u'time': duration,
+                    u'time_per_hit': duration / hits if hits else 0,
+                    u'percent': percent(duration, total_time),
+                    u'line': (line or u'').rstrip(),
                 }, file=out)
                 for (
                     _,
@@ -680,13 +800,13 @@ class ProfileBase(object):
                     callee_file, callee,
                 ) in call_list_by_line.get(lineno, ()):
                     print(_ANNOTATE_CALL_FORMAT % {
-                        'hits': call_hits,
-                        'time': call_duration,
-                        'time_per_hit': call_duration / call_hits,
-                        'percent': percent(call_duration, total_time),
-                        'callee_file': callee_file,
-                        'callee_line': callee.co_firstlineno,
-                        'callee_name': callee.co_name,
+                        u'hits': call_hits,
+                        u'time': call_duration,
+                        u'time_per_hit': call_duration / call_hits,
+                        u'percent': percent(call_duration, total_time),
+                        u'callee_file': callee_file,
+                        u'callee_line': callee.co_firstlineno,
+                        u'callee_name': callee.co_name,
                     }, file=out)
 
     def _iterRawFile(self, name):
@@ -971,7 +1091,7 @@ class Profile(ProfileBase, ProfileRunnerBase):
     # profile/cProfile-like API
     def run(self, cmd):
         """Similar to profile.Profile.run ."""
-        import __main__
+        from . import __main__
         dikt = __main__.__dict__
         return self.runctx(cmd, dikt, dikt)
 
@@ -1139,7 +1259,7 @@ class StatisticThread(threading.Thread, ProfileRunnerBase):
         stop_event = self._stop_event
         wait = partial(stop_event.wait, self._period)
         while self._can_run:
-            for ident, frame in current_frames().items():
+            for ident, frame in iterframes(current_frames=current_frames):
                 if test(ident):
                     sample(frame)
             frame = None
@@ -1220,10 +1340,13 @@ def _relpath(name):
     """
     return os.path.normpath(os.path.splitdrive(name)[1]).lstrip(_allsep)
 
-def _main(argv, stdin=None):
+def main(argv=None, stdin=None):
+    if argv is None:
+        argv = sys.argv
     format_dict = {
         'text': 'annotate',
         'callgrind': 'callgrind',
+        'callgrindzip': 'getCallgrindZip',
     }
 
     parser = argparse.ArgumentParser(argv[0])
@@ -1298,7 +1421,7 @@ def _main(argv, stdin=None):
         if options.script is not None:
             args.append(options.script)
         args.extend(options.argv)
-        import __main__
+        from . import __main__
         runner_method_kw = {
             'fd': stdin,
             'argv': args,
@@ -1317,7 +1440,9 @@ def _main(argv, stdin=None):
         }
         runner_method_id = 'runmodule'
     if options.format is None:
-        if _isCallgrindName(options.out):
+        if os.path.splitext(options.out)[1] == os.path.extsep + 'zip':
+            options.format = 'callgrindzip'
+        elif _isCallgrindName(options.out):
             options.format = 'callgrind'
         else:
             options.format = 'text'
@@ -1342,7 +1467,10 @@ def _main(argv, stdin=None):
             out = EncodeOrReplaceWriter(sys.stdout)
             close = lambda: None
         else:
-            out = io.open(options.out, 'w', errors='replace')
+            if options.format == 'callgrindzip':
+                out = io.open(options.out, 'wb')
+            else:
+                out = io.open(options.out, 'w', errors='replace')
             close = out.close
         if options.exclude:
             exclusion_search_list = [
@@ -1401,7 +1529,7 @@ def pprofile(line, cell=None):
     if cell is None:
         # TODO: detect and use arguments (statistical profiling, ...) ?
         return run(line)
-    return _main(
+    return main(
         ['%%pprofile', '-m', '-'] + shlex.split(line),
         io.StringIO(cell),
     )
@@ -1412,12 +1540,6 @@ except Exception:
     pass
 del pprofile
 
-def main():
-    _main(sys.argv)
-
-if __name__ == '__main__':
-    main()
-
 from ._version import get_versions
 __version__ = get_versions()['version']
 del get_versions
